/*  $Id: pion_server.pl, v1.3.0 2006/4/5  zhisheng Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2004-2006 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/

% this server requires XDIG version 1.5.0 or higher.

user:file_search_path(xdig, '../xdig').


:-use_module(xdig('dig_client')).
:-use_module(xdig('dig_server')).
:-use_module(xdig('dig_process')).
:-use_module(xdig('owl/owl2dig')).
:-use_module(xdig('dig_db')).

:-use_module(library('http/thread_httpd')).

:-use_module(xdig('pion/pion_process')).
:-use_module(xdig('pion/pion_selection')).


	



%deal with the tell request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=tells,
			!,
			dig_post(RequestData, Answer, [connection(close)]),
                  pion_setting(kb, KBID),
                  dig_assert_data(KBID, Data).



%deal with the ask request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=asks,
			!,
			dig_post(RequestData, Answer1, [connection(close)]),
			pion_answer_preprocessing(Answer1, Data, Answer),
                  true.


%deal with owl tell request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, _Data, Type),
			Type = 'rdf:RDF',
                  !,
               	owl2dig(RequestData, elements(L)),
			dig_post(elements(L), Answer, [connection(close)]),
                  dig_get_element(L, E, _),
                  E = element(tells, _, L1),
                 	pion_setting(kb, KBID),
 		     	dig_assert_data(KBID, L1),	
			true.


%deal with the set request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = set,
			!,
			dig_setting_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  true.

%deal with the do operation
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = do,
			!,
			dig_do_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  true.



%deal with other request
 my_dig_server_processing(RequestData, Answer, Options):-
			dig_post(RequestData, Answer, Options).


%pion_answer_preprocessing(+Answer, +Queries, -NewAnswer):-
%working on the Answer from the server with respect to the Queries
%if the current KB is inconsistent, then do inconsistency processing
%to obtain the NewAnswer, otherwise use the current Answer.


pion_answer_preprocessing(Answer1, Queries, Answer):-
    Answer1 = answer(_Header, AnswerBody),
    (AnswerBody = [element(responses, _, Elements)|_];AnswerBody = [element(response, _, Elements)|_]),
    pion_setting(kb, KBID),
    (pattern_set(Elements, incoherent); (dig_consistent_check(kb(KBID), Status),Status=false)),
    !,
    pion_setting(strategy, Strategy),
    pion_inconsistency_processing(KBID, Queries, Elements, Strategy, Answer).

pion_answer_preprocessing(Answer1, _Queries, Answer):-
		Answer = Answer1.

  			
%pion_inconsistency_processing(+KBID, +Queries, +AnswerFromServer, +Strategy, -MyAnswer)

pion_inconsistency_processing(KBID, Queries, _As, Strategy, Answer):- 
    pion_separation(Queries, Qs1, Qs2),
    pion_queries_answers(KBID, Qs1, Strategy, As1),
    dig_db_post(KBID, withClearKB, _, [connection(close)]),
    dig_ask(elements(Qs2), Answer1, [connection(close)]),
    dig_get_response_data(Answer1, As2),
%%  format(user_error, 'As1:~w~n As2:~w~n', [As1, As2]),
    append(As1, As2, As),
    !,
    dig_response_elements(As, Answer),
    dig_db_post(KBID, withClearKB, _, [connection(close)]).


pion_separation([], [], []):-!. 

pion_separation([H|T], [H|T1], T2):-
     H = element(Type, _, _),
     (Type = instance; Type = subsumes),
     !,
     pion_separation(T, T1, T2).

pion_separation([H|T], T1, [H|T2]):-!,
     pion_separation(T, T1, T2).





pion_queries_answers(_KBID, [], _, []):-!.

pion_queries_answers(KBID, [H|T], Strategy,  [A|As]):-
%    format(user_error, 'processing ~w~n', [H]),
    pion_query_answer(KBID, H, Strategy, A),
    !,
    pion_queries_answers(KBID, T, Strategy, As).




pion_query_answer(KBID, Q, Strategy, AnswerElement):-
       Q = element(instance, X, L),
       !,
      dig_get_element(L, element(individual, [name=I], []), L1),
      dig_get_element(L1, E, _),
      dig_get_parentconcept(KBID, I, L2),
      Q1 = element(subsumes, X, [E,L2]),   
	pion_query_answer(KBID, Q1, Strategy, AnswerElement),
      true.


pion_query_answer(KBID, Q, Strategy, A):-
     Q = element(subsumes, [], _L),      
      !,
	query_processing(Q, nil, KBID, Strategy, A),
      true.

pion_query_answer(KBID, Q, Strategy, A):-
     Q = element(subsumes, [id=ID], _L), 
      !,
	query_processing(Q, ID, KBID, Strategy, A), 
      true.
     
pion_query_answer(_KBID, Q, _, A):-
       Q = element(_, _, _),
       !,
       A = element(error, [unsupportedQueryFormat = Q],[]).



dig_get_parentconcept(KBID, I, E):-
   dig_db_element(KBID, element(instanceof, _, L)),
   dig_get_element(L, element(individual, [name=I],[]), L2),
   dig_get_element(L2, E, _),
   \+ E = element(top, [],[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% the start of  the linear extension strategy, 
%%%  version 1.2.4 renewed
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


query_processing(Q, ID, KBID, Strategy, A):-
          Strategy = linear_extension,
          !,
          pion_setting(selectionFunction, Type), 
	    linear_query_processing(Type, Q, ID, 1, [], icr([],[],[]), KBID, A).


query_processing(_Q, _ID, _KBID, Strategy, A):-
        A = element(error, [unsupportedStrategy = Strategy],[]).



linear_query_processing(Type, Q, ID, K, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K, P, ICRp, N, _L, _ICR, Status1),
     Status1 =  false,
     !,
     dig_db_element(setting(pion, odp),  Strategy),
     K1 is K+1,
     overdetermined_processing(Strategy, [P, Q, N], [Status2, L2, K2]),
     explain_answer(Status2, ID, L2, K1, K2, A).




linear_query_processing(Type, Q, ID, K, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K, P, ICRp, N, L, _ICR, _Status1),
     N = [],
     query_post(L, Q, Status2), 
     !,
     A = element(Status2, [id=ID, k=K+0],[]).




linear_query_processing(Type, Q, ID, K, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K, P, ICRp, _N, L, ICR, Status1),
     Status1 =  true,
     query_post(L, Q, Status2), 
     !,
     K1 is K+1,
     linear_query_processing2(Type, Q, ID, K1, KBID, L, ICR, Status2, A).



linear_query_processing(Type, Q, ID, K, P, ICRp, KBID, A):-
     pion_selection_function(Type, KBID, Q, K, P, ICRp, _N, L, ICR, Status1),
     Status1 =  true,
     query_post(L, Q, Status2), 
     !,
     linear_query_processing2(Type, Q, ID, K, KBID, L, ICR, Status2, A).


   
linear_query_processing(_Type, _Q, ID, _K, _P, _ICRp, _KBID, A):-
     A = element(error, [id=ID], []).



linear_query_processing2(Type, Q, ID, K1, KBID, L, ICR, Status2, A):-
      Status2 = undetermined,
      !,
      linear_query_processing(Type, Q, ID, K1, L, ICR, KBID, A).


linear_query_processing2(_Type, _Q, ID, K, _KBID, L, _ICR, Status2, A):-
(Status2 = accepted; Status2=rejected),
     !,
     explain_answer(Status2, ID, L, K, 0,  A).


linear_query_processing2(_Type, Q, ID, K1, _KBID, _L, _ICR, Status2, A):-
      format(user_error, 'do not know how to deal it:~w, Status2:~w~n', [Q, Status2]),
      A = element(error, [id=ID, k= K1],[]).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% the end of  the linear extension strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


overdetermined_processing(first_maximal_coherent_set, [P, Q, L], [Status1, L2, K]):-
   !,
   retract(pion_log_counter(odp,_)),
   assert(pion_log_counter(odp, 0)),
   overdetermined_processing(P, Q, L, Status1, L2, K).


% first_maximal_coherent_set as the default strategy.
overdetermined_processing(_Strategy, [P, Q, L], [Status1, L2, K]):-
   overdetermined_processing(P, Q, L, Status1, L2, K).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% overdetermined processing
%%% first_maximal_coherent_set_strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%overdetermined_processing(+WorkingSet, +Query,  +CurrentSelectedSet, -Status, -NewWorkingSet)

overdetermined_processing(_L1, _Q, _L, Status, _L2, K):-
	pion_log_counter(odp, I),
      dig_db_element(setting(pion, odpLimit), IL),
      I > IL,
      !,
      Status = undetermined,
      K = IL.


% search for the maximal conherent set by deleting one element
overdetermined_processing(L1, Q, L, Status1, L2, K):-
            select(_E, L, H),
   %        show_overdetermined_processing(E), % for debugging only
            union(L1, H, L2),
            query_post(L2, Q, Status),
     		pion_log_counter(odp, I),
            I1 is I+1,
            retract(pion_log_counter(odp, I)),
            assert(pion_log_counter(odp, I1)),
   %         format(user_error, 'Q:~w, Status:~w, ignoring ~w~n', [Q, Status, E]),
		(Status = accepted; Status = rejected; Status= undetermined),           
            !,
            Status1 = Status,
            K = I1,
            true.

% if deleting one element is not enough to get a consistent set,
% then deleting one more to continue the processing.

overdetermined_processing(L1, Q, L, Status1, L2, K1):-
            select(_E, L, H),
            H \= [],
            !,
            overdetermined_processing(L1, Q, H, Status, L2, K),
            K1 is K +1,
            Status1 = Status.



overdetermined_processing(_L1, _Q, _, Status, [element(message,['fail to find relevant consistent set'], [])], K):-
         !,
	   Status = undetermined,
         K =1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% end of first_maximal_coherent_set_strategy
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



query_post(L, Q, Status):-
      !,
      L1 =[element(clearKB, [], [])|L],
	dig_tell(elements(L1), answer(_, _Answer1), [connection(close)]),
 %%    message_log(L1), % for log only      
	dig_ask(elements([Q]), answer(_, Answer2), [connection(close)]),   
	opposite_query(Q, Q1),
	dig_ask(elements([Q1]), answer(_, Answer3), [connection(close)]),  
      answer_status(Answer2, Answer3, Status),  
	true.


opposite_query(Q, Q1):-
      Q = element(instance, _X, L),
	!,
      dig_get_element(L, I, L1),
      dig_get_element(L1, C, _L2),
      dig_add_elements([C], not, C1),
      Q1= element(instance, [], [I|C1]).


opposite_query(Q, Q1):-
      Q = element(subsumes, _X, L),
      !,
      dig_get_element(L, C1, L1),
      dig_get_element(L1, C2, _L2),
      C3 = element(not, [], [C1]),
      Q1= element(subsumes, [], [C3, C2]).



opposite_query(Q, Q1):-
      dig_add_elements([Q], not, [Q1]).


%explain_answer(+Status, +ID, +CurrentWorkingList, +K, +K1, -Answer)  %% version 1.2.4 added

explain_answer(Status, ID, L, K, K1, Answer):-
	(Status = accepted; Status = rejected),
      pion_setting(explain, on),
      !,
      delete_repeated_element(L, L1),
     	Answer = element(Status,[id=ID, k=K+K1], [element(because, [], L1)]).

explain_answer(Status, ID, _L, K, K1, Answer):-
     Answer = element(Status, [id=ID, k=K+K1],[]).



%explain_answer(+Status, +ID, +CurrentWorkingList, -Answer)

explain_answer(Status, ID, L, Answer):-
	(Status = accepted; Status = rejected),
      pion_setting(explain, on),
      !,
      delete_repeated_element(L, L1),
     	Answer = element(Status,[id=ID], [element(because, [], L1)]).


explain_answer(Status, ID, _L, Answer):-
     Answer = element(Status, [id=ID],[]).

delete_repeated_element([], []):-!.

delete_repeated_element([X|Y], Z):-member(X, Y),
                        !,
                        delete_repeated_element(Y, Z).

delete_repeated_element([X|Y], [X|Z]):-
                        delete_repeated_element(Y, Z).




%the +answer set has an element with 
%the pattern of the message (from the server)
%define the pattern of the incoherency message for RACER

%it may not work for other untested DIG system.

pattern_set(As, incoherent):-
    E=element(error, L, _),
    member(E, As),
    dig_select_attribute(L, message, M),
    sub_string(M, _S, _L, _B, 'incoherent').

%The  +Element has the pattern
%define the pattern of the incoherency message for RACER

pattern(E, incoherent):-
    E=element(error, L, _),
    dig_select_attribute(L, message, M),
    sub_string(M, _S, _L, _B, 'incoherent').



pion_setting(X, Y):-
   dig_db_element(setting(pion, X), Y).

pion_init:-
   retractall(dig_db_element(setting(pion, _), _)), 
%  assert(dig_db_element(setting(pion, selectionFunction), syntactic_relevance)),
   assert(dig_db_element(setting(pion, selectionFunction), syntactic_concept_relevance)),
   assert(dig_db_element(setting(pion, strategy), linear_extension)),
   assert(dig_db_element(setting(pion, kb), general)),
   assert(pion_log_counter(0)),
   assert(dig_db_element(setting(pion, odp), first_maximal_coherent_set)),
   assert(dig_db_element(setting(pion, odpLimit), 1000)),
   assert(pion_log_counter(odp, 0)),
   assert(dig_db_element(setting(pion, explain), off)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% predicates for debugging and log only
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%/*
show_overdetermined_processing(E):-
          E = element(Type, _, EL),            
            E1 = element(_, [name=N], _),
            dig_get_element(EL, E1, _),
            format(user_error, 'Ignoring: axiom(~w_~w)~n~n', [Type, N]).


message_log(_L):-
      pion_log_counter(I),
	I > 500,  % limit of the log counter
      !.

message_log(L):-
      pion_log_counter(I),
      I1 is I+1,
      retract(pion_log_counter(I)),
      assert(pion_log_counter(I1)),
      concat_atom(['log/tmp_tell_log', I, '.xml'], File0),
      !,
      open(File0, write, Out, []),
	dig_xmlns(XMLNS),
	TellElements = [element(tells, XMLNS, L)],
      elements_xmltext(TellElements, file(File0)),
 	close(Out),
      true.

message_log(_L):-!.

%*/
   

			

:- pion_init, dig_server_port(Port), set_dig_server_id('SEKT PION Server 1.2.4'), http_server(dig_server,[port(Port)]).

     
