/*  $Id: dig_client.pl, v1.5.0 2006/4/5  zhisheng Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2004-2006 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/


:- module(dig_client,
	  [ dig_post/3,				% +File, -Reply, +Options
	    dig_post/4,				% +URL, +File, -Reply, +Options
	    dig_post/7, 			% +DigProtocol, +DigHost, +DigPort, +DigPath, +File, -Reply, +Options
	    dig_tell/3,				% +Elements, -Answer, +Options
	    dig_ask/3				% +Elements, -Answer, +Options
	  ]).


:-use_module(xdig('dig_client_setting')).
:-use_module(xdig('dig_process')).
:-use_module(library('http/http_client')).

:-use_module(library('memfile')).




% dig_post(+URL, +File, answer(-Header,-XmlElements), +Options)


dig_post(URL, elements(Elements), answer(Header, XmlElements), Options):-!,
	dig_tmp_working_file(xml, XMLFile),
	elements_xmltext(Elements, file(XMLFile)),
	http_post(URL, file(XMLFile), RawTextReply, Options),
	xml_elements(RawTextReply, XmlElements, Header),
	true.

/* dig_post(URL, elements(Elements), Reply, Options):-!,
	elements_xmltext(Elements, text(XMLText)),
        dig_post(URL, text(XMLText), Reply, Options),
	true.

*/




dig_post(URL, text(XMLText), answer(Header, XmlElements), Options):-!,
	new_memory_file(MemoryFile),
	open_memory_file(MemoryFile, write, Out),
	format(Out, 'Content-type: text/xml\n\n', []),
	write(Out, XMLText),
	close(Out),
	size_memory_file(MemoryFile, Len),
	open_memory_file(MemoryFile, read, In),
        http_post(URL, cgi_stream(In, Len), RawTextReply, Options), 
        free_memory_file(MemoryFile),
	xml_elements(RawTextReply, XmlElements, Header), 
	true.
	


dig_post(URL, File, answer(Header, XmlElements), Options):-
	http_post(URL, File, RawTextReply, Options),
	xml_elements(RawTextReply, XmlElements, Header).


%dig_post(+DigProtocol, +DigHost, +DigPort, +In, -Answer, +Options)

dig_post(P, DigHost, DigPort, DigPath, File, Answer, Options):-
	URL = [protocol(P),host(DigHost), port(DigPort), path(DigPath)],
	dig_post(URL, File, Answer, Options).

dig_post(File, Reply, Options):-
	dig_url(URL),
	dig_post(URL, File, Reply, Options),
        true.



dig_post_status(URL, File, Answer, Options, Status):-
	dig_post(URL, File, answer(Header, XmlElements), Options),
	Answer = answer(Header, XmlElements),
	get_header_value(Header, reply_code, Status).



dig_tell(elements(Elements), Answer, Options):-
			dig_add_elements(Elements, tells, TellElements),
			dig_post(elements(TellElements), Answer, Options),
			true.

		
dig_ask(elements(Elements), Answer, Options):-
			dig_add_elements(Elements, asks, AskElements),
			dig_post(elements(AskElements), Answer, Options),
			true.







