/*  $Id: dig_process.pl, v1.5.0 2006/4/5  zhisheng Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2004-2006 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/


:- module(dig_process,
	  [ dig_add_elements/3,			% +Data, +Type, -ElementsWithType
	    xml_elements/3,			% +RawText, -XMLElements, -Header
	    elements_xmltext/2,			% +XMLElements, -XMLText
	    elements_xmltext_text/2,		% +XMLElements, -XMLText  %% version 1.1
	    xmltext_elements/2,			% +XMLText, -XMLElements, %% version 1.1
	    dig_requestdata_analysis/3,	%+RequestData, -Data, -Type
	    dig_data_analysis/4,		%+Data, +Type, -Element, -OtherData
	    dig_response_analysis/4,		%+Data, +Type, -Element, -OtherData
	    dig_get_response_data/2,		%+Answer, -Responsedata, %% version 1.4.6
	    dig_select_attribute/3, 		% +AttributeList, +Type, -Message, %% version 1.1
	    dig_get_element/3,			% +List, +Element, -Other,   %% version 1.1
	    dig_xmlns/1,				%-XMLNS List
	    dig_all_conceptname/2,		%+KBID, -AllConceptNames, %% version 1.4.2
	    dig_all_individual/2,		%+KBID, -IndividualList, %% version 1.4.2
	    dig_all_role/2,			%+KBID, -RoleList, %% version 1.4.4
	    dig_unsatisfiable_conceptname/2,	%+KBID, -AllUnsatisfiableConceptName, %% version 1.4.2
	    dig_unsatisfiable_conceptname/1,	%-AllUnsatisfiableConceptName, %% version 1.4.2
            dig_unsatisfiable_conceptname/3,	%+KBID or Elements, -UnsatisfiableConcepts, -Status, %% version 1.4.6
	    dig_related_conceptname/4,		%+KBID, +Concept, +Type, -RelatedConcepts, %% version 1.4.2
	    dig_related_conceptname/3,		%+Concept, +Type, -RelatedConcepts, %% version 1.4.6
            allconceptname/2,			%+ConceptList, -ConceptNameLists, %% version 1.4.6
            allindividualname/2,	     	%+IndividualList, -IndividualNameLists, %% version 1.4.6
            allrolename/2,			%+RoleList, -RoleNameLists, %% version 1.4.6
	    dig_satisfiable_concept_check/3,%KB or Elements, +Concept, -Status, %% version 1.4.6
          dig_consistent_check/2,		%KB or Elements, -Status, %% version 1.4.6 
	    dig_tmp_working_file/2		% +Type, -File
	  ]).



:-use_module(library('memfile')).	%% internal memory file
:-use_module(library('sgml')).

:-use_module(library('sgml_write')).


:- dynamic
	tmp_working_file/2.

dig_get_response_data(Answer, Data):-
     Answer = answer(_, L),
     dig_get_element(L, E, _),
     E = element(_,_, Data).




xml_elements(RawText, Elements, HDRText) :-
	atom_codes(RawText, RawCodes),
	skip_until_xml(RawCodes, XMLCodes, HDRCodes),
	atom_codes(XMLText, XMLCodes),
	atom_codes(HDRText, HDRCodes),
	xmltext_elements(XMLText, Elements).


xmltext_elements(XMLText, Elements):-
	atom_to_memory_file(XMLText, MemoryHandle),
	open_memory_file(MemoryHandle, read, MemoryStream),
	load_structure(MemoryStream, Elements, [dialect(xml)]),    
	close(MemoryStream),
	free_memory_file(MemoryHandle).



skip_until_xml([], [], []).
skip_until_xml(CI, CO, []) :-
	CI = [0'<, 0'?, 0'x, 0'm, 0'l | _],
	!,
	keep_xml_codes(CI, CO).

%'

skip_until_xml([H|T], CO, [H|Next]) :-
	skip_until_xml(T, CO, Next).

keep_xml_codes([], []) :-
	!.
keep_xml_codes(CI, CO) :-
	CI = [0'<, 0'/, 0'r, 0'e, 0's, 0'p, 0'o, 0'n, 0's, 0'e, 0'> | _],
	CO = [0'<, 0'/, 0'r, 0'e, 0's, 0'p, 0'o, 0'n, 0's, 0'e, 0'> ],
	!.
keep_xml_codes([HI|TI], [HI|TO]) :-
	keep_xml_codes(TI, TO).


get_header_value([], reply_code, code('cannot find the http head')):-!.
get_header_value(CI, reply_code, code(Code)):-
		CI = [0'H, 0'T, 0'T, 0'P, 0'/, _Major, 0'., _Minor, _Space, Code1, Code2, Code3 | _],
		!,
		CodeList = [Code1, Code2, Code3],
            atom_codes(Code, CodeList).

get_header_value([_H|T], reply_code, Code):-
		get_header_value(T, reply_code, Code).

% elements_xmltext(+Elements, XMLText)




elements_xmltext(Elements, file(XMLFile)):-
	open(XMLFile, write, Out, []),
%        format(user_error, 'Elements=~q~n', [Elements]),
	xml_write(Out, Elements, []),
	close(Out),
	true.

elements_xmltext(Elements, stream(Out)):-
	xml_write(Out, Elements, []).


/*
elemets_xmltext(Elements, text(Text)):-
	dig_tmp_working_file(xml2, XMLFile),
        open(XMLFile, write, Out, []),
	xml_write(Out, Elements, []),
	close(Out),
	open(XMLFile, read, In, []),
        read(In, Text),
        close(in).
*/

        

%% version 1.1
elements_xmltext(Elements, text(Text)):-
       elements_xml_text(Elements, Text).
	


print_xml_head(Out):-
		write(Out, '<?xml version="1.0" encoding="ISO-8859-1"?>').


elements_xml(L,  Out):-
   elements_xml_text(L, Text),
   format(Out, '~w', [Text]).





        

dig_add_elements(Elements, tells, TellElements):-
                   !,
			dig_xmlns(XMLNS),
			TellElements = [element(tells, XMLNS, Elements)].

dig_add_elements(Elements, asks, TellElements):-
                  !,
			dig_xmlns(XMLNS),
			TellElements = [element(asks, XMLNS, Elements)].

dig_add_elements(Elements, clearKB, NewElements):-
                  !,
			NewElements = [element(clearKB, [],[])|Elements].

dig_add_elements(Elements, Type, NewElements):-
			NewElements = [element(Type, [], Elements)].



dig_requestdata_analysis(RequestData, Data, Type):-
			RequestData = elements(Elements),
			Elements = [element(Type, _, Data)].


dig_requestdata_analysis(RequestData, Data, Type):-
			RequestData = text(XMLText),
			xml_elements(XMLText, Elements, _Header),
			Elements = [element(Type, _, Data)].


dig_requestdata_analysis(RequestData, RequestData, tell):-!.

dig_data_analysis(Data, Type, Element, OtherData):-
		Element =element(Type, _, _),
		select(Element, Data, OtherData).


dig_data_analysis(Data, _Type, nil, Data):-!.

dig_response_analysis(AnswerBody, Type, Element, Other):-
			AnswerBody = [element(response, _, Elements)|_],
			Element=element(Type,_,_),
			select(Element, Elements, Other).

dig_response_analysis(AnswerBody, Type, Element, Other):-
			AnswerBody = [element(responses, _, Elements)|_],
			Element=element(Type,_,_),
			select(Element, Elements, Other).

dig_response_analysis(_,_,nil,_):-!.


dig_tmp_working_file(txt, 'tmp_xml.txt').
dig_tmp_working_file(xml, 'tmp_xml.xml').
dig_tmp_working_file(xml0, 'tmp_xml0.xml').
dig_tmp_working_file(xml1, 'tmp_xml1.xml').
dig_tmp_working_file(xml2, 'tmp_xml2.xml').

dig_xmlns([xmlns='http://dl.kr.org/dig/lang', xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance', xsi:schemaLocation='http://wasp.cs.vu.nl/sekt/dig/dig-1.1.xsd']):-!.



%% version 1.1
dig_get_element([], nil, []):-!.
dig_get_element([E|T], E1, T):-
     E= element(_,_,_),
     !,
     E1 = E.

dig_get_element([_E|T],E,L):-
     !,
     dig_get_element(T,E,L).




dig_select_attribute([],_, nil):-!.
dig_select_attribute([Type=M|_T], Type, M):-!.
dig_select_attribute([_|T], Type, M):-
   dig_select_attribute(T, Type, M).

elements_xmltext_text(Elements, Text):-
		print_xml_head_text(Text1),
		elements_xml_text(Elements, Text2),
            concat_atom([Text1, Text2], Text).


	

print_xml_head_text(Text):-
		sformat(Text, '<?xml version="1.0" encoding="UTF-8"?>', []).



elements_xml_text([], ' '):-!.

elements_xml_text([H|T], Text):-
	element_xml_text(H, Text1),
	!,
	elements_xml_text(T, Text2),
      concat_atom([Text1, Text2], Text).




element_xml_text(element(Node, [], []), Text):-!,
       sformat(Text, '<~w/>',  [Node]).

element_xml_text(element(Node, Attrib, []), Text):-!,
       sformat(Text1, '<~w',  [Node]),
       print_all_attributes_text(Text2, Attrib),
	 sformat(Text3, '/>', []),
       concat_atom([Text1, Text2, Text3], Text).
	 
		
element_xml_text(element(Node, Attributes, Elements), Text):-!,
       sformat(Text1, '<~w',  [Node]),
	 print_all_attributes_text(Text2, Attributes),
       sformat(Text3, '>',[]),
	 elements_xml_text(Elements, Text4),
	 sformat(Text5, '</~w>', [Node]),
       concat_atom([Text1, Text2, Text3, Text4, Text5],Text).
       

element_xml_text(X, Text):-
     delete_code(X, 13, Text1),
     delete_code(Text1, 10, Text).

element_xml_text(X, X):-
		!.
% capture all statement to avoid the error in the predicate.
% new added for version 1.4

		
     
print_all_attributes_text('', []):-!.
print_all_attributes_text(Text, [H=Value|T]):-
		sformat(Text1, ' ~w = "~w"', [H, Value]),
		!,
		print_all_attributes_text(Text2, T),
            concat_atom([Text1, Text2], Text).


print_all_attributes_text(Text, [H|T]):-
		sformat(Text1, ' ~q', [H]),
		!,
		print_all_attributes_text(Text2, T),
            concat_atom([Text1, Text2], Text).



delete_code(Text, Code, Newtext):-
		atom_codes(Text, Codes),
		delete(Codes, Code, NewCodes),
		atom_codes(Newtext, NewCodes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.4.2 added, version 1.4.6 revised
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% dig_all_conceptname(+KBID, -AllConceptNames)

dig_all_conceptname(ID, L):-
     dig_db_post(ID, withClearKB, _Answer, [connection(close)]),
     dig_all_conceptname(L).


 dig_all_conceptname(L2):-
     dig_text(allconcept, Value),
     dig_post(text(Value), Reply, [connection(close)]),
     Reply = answer(_, L),
     L = [element(_,_, L1)],
     select(element(conceptSet, [id=_], L2), L1, _L3).


dig_all_individual(ID, L):-
     dig_db_post(ID, withClearKB, _Answer, [connection(close)]),
     dig_all_individual(L).

 dig_all_individual(L8):-    
     dig_text(allindividual, Value2),
     dig_post(text(Value2), Reply2, [connection(close)]),
     Reply2 = answer(_, L6),
     L6 = [element(_,_, L7)],
     select(element(individualSet, [id=_], L8), L7, _L9),   
 %   allindividualname(L8, L10),
     true.

%dig_unsatisfiable_conceptname(+KBID, -AllUnsatisfiableConceptName)

dig_unsatisfiable_conceptname(ID, L):-
     dig_db_post(ID, withClearKB, _Answer, [connection(close)]),
     dig_unsatisfiable_conceptname(L),
     true.

%dig_unsatisfiable_conceptname(-AllUnsatisfiableConceptName)
% with respect to the told data in the external DL reasoner

dig_unsatisfiable_conceptname(L5):-
     dig_text(allconcept, Value),
     dig_post(text(Value), Reply, [connection(close)]), 
     Reply = answer(_, L),
     L = [element(_,_, L1)],
     select(element(conceptSet, [id=_ID], L2), L1, _L3), 
     unsatisfiable_conceptname(L2, L5).


%unsatisfiable_conceptname(+UnsatisfiableConceptNameListInDIGResponseFormat, -UnsatisifiableConceptNames)

unsatisfiable_conceptname(L1, L):-
    select(element(synonyms, _, L2), L1, L3),
    member(element(bottom,[],[]), L2),
    !,
    select(element(bottom, [],[]), L2, L21),
    unsatisfiable_conceptname(L3, L4),
    append(L21, L4, L).

unsatisfiable_conceptname(L1, L4):-
    select(element(synonyms, _, _L2), L1, L3),
    !,
    unsatisfiable_conceptname(L3, L4).

unsatisfiable_conceptname(_, []):-!.


allconceptname(L1, L):-
    select(element(synonyms, _, L2), L1, L3),
    !,
    allconceptname(L3, L4),
    append(L2, L4, L).


allconceptname(_, []):-!.


allrolename(L1, L):-
	allconceptname(L1, L).

allindividualname(L1, L):-
    select(element(individual, [name=I], _), L1, L3),
    sub_string(I, X, _Y, _Z, ':'),
    X > 0,
    !,
    allindividualname(L3, L).

allindividualname(L1, L):-
    select(element(individual, [name=I], _), L1, L3),
    !,
    allindividualname(L3, L4),
    append([I], L4, L).


allindividualname(_, []):-!.





get_shorted_concept_name(C2, C3):-
    sub_string(C2, X, _Y, Z, '#'),
    X>0,
    !,
    X1 is X+1,
    sub_string(C2, X1, Z, _W, C3).


get_shorted_concept_name(C2, C2):-!.


dig_text(allconcept, Value):-
Value = '<?xml version="1.0" encoding="ISO-8859-1" ?><asks xmlns="http://dl.kr.org/dig/lang" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://dl.kr.org/dig/lang wasp.cs.vu.nl/sekt/digtest/dig-1.1.xsd"><allConceptNames/></asks>'.


 
dig_text(allindividual, Value):-
Value = '<?xml version="1.0" encoding="ISO-8859-1" ?><asks xmlns="http://dl.kr.org/dig/lang" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://dl.kr.org/dig/lang wasp.cs.vu.nl/sekt/digtest/dig-1.1.xsd"><allIndividuals/></asks>'.


dig_text(allrole, Value):-
Value = '<?xml version="1.0" encoding="ISO-8859-1" ?><asks xmlns="http://dl.kr.org/dig/lang" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://dl.kr.org/dig/lang wasp.cs.vu.nl/sekt/digtest/dig-1.1.xsd"><allRoleNames/></asks>'.

dig_text(Type, C, Value):-!,
Head = '<?xml version="1.0" encoding="ISO-8859-1" ?><asks xmlns="http://dl.kr.org/dig/lang" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://dl.kr.org/dig/lang wasp.cs.vu.nl/sekt/digtest/dig-1.1.xsd">',
sformat(Text1, '<~w id="~w"><catom name="~w"/></~w>', [Type, C, C, Type]),
concat_atom([Head, Text1, '</asks>'], Value).


% dig_related_conceptname(+KBID, +Concept, +Type, -RelatedConcepts)
% Type = parents, children, ancestors, descendants, or equivalents

dig_related_conceptname(ID, C, Type, L):-
     dig_db_post(ID, withClearKB, _Answer, [connection(close)]),
     dig_related_conceptname(C, Type, L).

% deal with normal case
 dig_related_conceptname(C, Type, L5):-
     dig_text(Type, C, Value),
     dig_post(text(Value), Reply, [connection(close)]),
     Reply = answer(_, L),
     L = [element(_,_, L1)],
     select(element(conceptSet, [id=_ID], L2), L1, _L3),
     !, 
     allconceptname(L2, L5).

% deal with the case Concept C does not appear in the ontology
dig_related_conceptname(_C, _Type, L5):-
 L5=[].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.4.4 added, version 1.4.6 revised
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   

dig_all_role(ID, L):-
     dig_db_post(ID, withClearKB, _Answer, [connection(close)]),
     dig_all_role(L).

 dig_all_role(L8):-    
     dig_text(allrole, Value2),
     dig_post(text(Value2), Reply2, [connection(close)]),
     Reply2 = answer(_, L6),
     L6 = [element(_,_, L7)],
     select(element(roleSet, [id=_], L8), L7, _L9).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  version 1.4.6 added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
dig_unsatisfiable_conceptname(KB, UC, Status):-
	KB =kb(KBID),
      !,
      dig_unsatisfiable_conceptname(KBID, UC),
	(member(element(catom, _,_),  UC)-> Status = false; Status =true),
      true.
     
dig_unsatisfiable_conceptname(KB, UC, Status):-
      KB = elements(L2),
      !,
      Options = [connection(close)],
      L3 = [element(clearKB, [], [])|L2],
      dig_tell(elements(L3), _Reply, Options),
      dig_unsatisfiable_conceptname(UC),
      (member(element(catom, _,_),  UC)-> Status = false; Status =true),
      true.

 dig_unsatisfiable_conceptname(KB, UC, Status):-   
      !,
      Status = error,
      UC = nil,
      format(user_error, 'never be here for satisfiable check for ~w!!!~n', [KB]).


%dig_satisfiable_concept_check(+KB or Elements, +Concept, -Status)

dig_satisfiable_concept_check(KB, C, Status):-
      dig_unsatisfiable_conceptname(KB, UC, _Status1),
      !,
      E = element(catom, [name=C], []),
      (member(E, UC)->Status = true; Status = false).


dig_consistent_check(KB, Status):-
     dig_unsatisfiable_conceptname(KB, _UC, Status),
 %   format(user_error, 'Status: ~w on ~w~n because unsatisfiable concet:~w~n~n', [Status, KB, UC]),
     true.



